/*
  Copyright 2007 Eurotech S.P.A. (Luca Calligaris)
  Copyright 2007 Christian Pellegrin <chripell@gmail.com> EVOL S.R.L.
  
  This software is released under the GPL v2. Please see the file
  COPYING in the top-level directory.
*/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <netinet/in.h>

#include "libe0904.h"

struct e0904_ch* simple_e0904_encoder(char *dev) {
  struct e0904_ch *e;
  void *m;
  int ret;

  e = malloc(sizeof(*e));
  assert(e);

  e->last_isNtsc = 666;

  e->fd[0] = open(dev, O_RDWR);
  if (e->fd[0] == -1) {
    perror("simple_e0904_encoder open");
    return NULL;
  }

  if ((ret = ioctl(e->fd[0], IOCTL_ETHJPG2K_LINUX_GETBSIZE, &e->bsize)) == -1) {
    perror("ioctl IOCTL_ETHJPG2K_LINUX_GETBSIZE");
    return NULL;
  }  
  e->bsize *= 4096;
  //printf("buffer size: %ld\n", e->bsize);
  
  m = mmap(NULL, 2 * e->bsize, PROT_READ|PROT_WRITE , MAP_SHARED, e->fd[0], 0);
  if (m == MAP_FAILED) {
    perror("mmap");
    return NULL;
  }
  e->b = (unsigned char *) m;

  e->bsize /= 4;

  return e;
}


int simple_e0904_board_init(struct e0904_ch *e, int isNtsc , char *b, 
			    ETHJPG2K_FIRMWARE_DATA *fwdata,
			    ETHJPG2K_FIRMWARE_PARAMS *fwparams,
			    int slave_path) {
  int ret;
  unsigned char firmware[32768];
  
  FILE *f;
  
  f = fopen(b, "r");
  if (!f) {
    perror("open firmware");
    fprintf(stderr, "cannot open firmware: %s\n", b);
    return -1;
  }
  fread(firmware, 32768, 1, f);
  fclose(f);

  fwdata->pFirmwareParameters = fwparams;
  fwdata->pFirmware = (unsigned long *) firmware;

  if ((ret = ioctl(e->fd[0], IOCTL_ETHJPG2K_ENCDEC_LOAD_FIRMWARE, fwdata)) == -1) {
    perror("ioctl IOCTL_ETHJPG2K_ENCDEC_LOAD_FIRMWARE");
    return ret;
  }

  if (isNtsc != e->last_isNtsc) {
    if ((ret = ioctl(e->fd[0], IOCTL_ETHJPG2K_LINUX_INIT, isNtsc)) == -1) {
      perror("ioctl IOCTL_ETHJPG2K_LINUX_INIT");
      return ret;
    }
    e->last_isNtsc = isNtsc;
  }

  if ((ret = ioctl(e->fd[0], IOCTL_ETHJPG2K_SET_SLAVE_PATH, &slave_path)) == -1) {
    perror("ioctl IOCTL_ETHJPG2K_SET_SLAVE_PATH");
    return ret;
  }

  e->cod_style = fwparams->DecodeParams.COD_STYLE;
  e->video_norm = fwparams->DecodeParams.VFORMAT;
  
  return 0;
}


int simple_e0904_encoder_run(struct e0904_ch* e, e0904_frame_cb cb) {
  int ret;
  struct ethjpg2k_linux_hdr hdr;  
  
  
  if ((ret = ioctl(e->fd[0], IOCTL_ETHJPG2K_ENCDEC_START)) == -1) {
    perror("ioctl IOCTL_ETHJPG2K_ENCDEC_START");
    free(e);
    return ret;
  }

  ret = 0;
  while(!ret) {
    ret = read(e->fd[0], &hdr, sizeof(hdr));
    if (ret < 0) {
      perror("read hdr");
      break;
    }
    ret = cb(&hdr, e->b + hdr.start, hdr.size*4);
    write(e->fd[0], &hdr.ticket, sizeof(unsigned long));
  }
  ioctl(e->fd[0], IOCTL_ETHJPG2K_ENCDEC_STOP);
  return ret;
}

int simple_e0904_decoder_run(struct e0904_ch* e, e0904_len_cb cbl, e0904_frame_in_cb cbf, int add_hdr) {
  int ret, imsize, hdr_size = 0;
  int count = 0, j;
  struct ethjpg2k_linux_decode dec;  

  if (add_hdr)
    hdr_size = 16;

  if ((ret = ioctl(e->fd[0], IOCTL_ETHJPG2K_ENCDEC_START)) == -1) {
    perror("ioctl IOCTL_ETHJPG2K_ENCDEC_START");
    return ret;
  }

  while(1) {
    ret = imsize = cbl();
    if (ret < 0)
      break;
    while (1) {
      ret = read(e->fd[0], &dec, sizeof(dec));
      if (ret < 0) {
	perror("decoder read");
	goto out_of_here;
      } else {
	if (imsize + hdr_size < dec.size) 
	  break;
	usleep(100 * 1000);
      }
    }
    cbf(e->b + dec.start + hdr_size, imsize);
    dec.size = imsize + hdr_size;
    if (add_hdr) {
      unsigned int *h = (unsigned int *) (e->b + dec.start);
      
      h[0] = 0xfffffff0 | (count & 1);
      h[1] = count;
      h[2] = (e->cod_style << 24) | (e->video_norm << 16) | 2;
      h[3] = imsize / 4;
      for(j=0; j < 4; j++) {
	h[j] = htonl(h[j]);
      }
    }
    ret = write(e->fd[0], &dec, sizeof(dec));
    if (ret < 0) {
      perror("decoder write");
      break;
    }
    count ++;
  }
 out_of_here:
  ioctl(e->fd[0], IOCTL_ETHJPG2K_ENCDEC_STOP);
  return ret;
}

void simple_e0904_force_stop(struct e0904_ch* e) {
  ioctl(e->fd[0], IOCTL_ETHJPG2K_ENCDEC_STOP);
}

int simple_e0904_ioctl(struct e0904_ch* e, int code, void *arg) {
  return ioctl(e->fd[0], code, arg);
}

void simple_e0904_free(struct e0904_ch* e) {
  close(e->fd[0]);
  free(e);
}

void simple_e0904_board_noinit(struct e0904_ch *e, int isNtsc) {
  e->last_isNtsc = isNtsc;
}

int simple_e0904_hipi(struct e0904_ch *enc, struct e0904_ch *dec,
		      char *fw, ETHJPG2K_FIRMWARE_DATA *fwdata, ETHJPG2K_FIRMWARE_PARAMS *fwparams,
		      int n, 
		      char **in, int *ins,
		      char **out, int *outs) {
  int ret=0;
  int i,j;
  struct ethjpg2k_linux_decode dech;  
  struct ethjpg2k_linux_hdr hdr;  
  unsigned char firmware[32768];  
  FILE *f;
  int ns = 2;			/* this accounts for HIPI being one frame behind */
  
  f = fopen(fw, "r");
  if (!f) {
    perror("open firmware");
    fprintf(stderr, "cannot open firmware: %s\n", fw);
    return -1;
  }
  fread(firmware, 32768, 1, f);
  fclose(f);

  fwdata->pFirmwareParameters = fwparams;
  fwdata->pFirmware = (unsigned long *) firmware;
  if ((ret = ioctl(enc->fd[0], IOCTL_ETHJPG2K_ENCDEC_LOAD_FIRMWARE, fwdata)) == -1) {
    perror("ioctl IOCTL_ETHJPG2K_ENCDEC_LOAD_FIRMWARE");
    return ret;
  }
  
  if ((ret = ioctl(dec->fd[0], IOCTL_ETHJPG2K_ENCDEC_START)) == -1) {
    perror("ioctl dec IOCTL_ETHJPG2K_ENCDEC_START");
    return ret;
  }
  if ((ret = ioctl(enc->fd[0], IOCTL_ETHJPG2K_ENCDEC_START)) == -1) {
    perror("ioctl enc IOCTL_ETHJPG2K_ENCDEC_START");
    return ret;
  }

  for(i=0; i<n;i++) {
    for (j=0; j<ns; j++) {
      while (1) {
	ret = read(dec->fd[0], &dech, sizeof(dech));
	if (ret < 0) {
	  perror("decoder read");
	} else {
	  if (ins[i] < dech.size) 
	    break;
	  usleep(100 * 1000);
	}
      }
      memcpy(&dec->b[dech.start], in[i], ins[i]);
      dech.size = ins[i];
      ret = write(dec->fd[0], &dech, sizeof(dech));
      if (ret < 0) {
	perror("decoder write");
	break;
      }
    }
    ns = 1;

    ret = read(enc->fd[0], &hdr, sizeof(hdr));
    if (ret < 0) {
      perror("encoder read hdr");
      break;
    }
    out[i] = malloc(hdr.size*4);
    assert(out[i]);
    outs[i] = hdr.size*4;
    memcpy(out[i], &enc->b[hdr.start], hdr.size*4);
    write(enc->fd[0], &hdr.ticket, sizeof(unsigned long));
  }
  ioctl(enc->fd[0], IOCTL_ETHJPG2K_ENCDEC_STOP);
  ioctl(dec->fd[0], IOCTL_ETHJPG2K_ENCDEC_STOP);
  return ret;
}

